from jaepeto.patterns.entrypoint import (
    get_script_entrypoints,
    group_imports_at_level,
    infer_important_functions,
)
from jaepeto.patterns.interactions import describe_file_architecture
from jaepeto.patterns.packages import PyVersionDetector, read_packages
from jaepeto.patterns.tools import CIParser, ContainerParser



"""
Third-party tools
e.g., Coverage, Docker
"""
import os
import re
from pathlib import Path
from typing import List, Optional

import requests

from jaepeto.utils import post


class ContainerParser:
    def __init__(self) -> None:
        self.file: Optional[str] = None

    def parse(self, project_dir: Path) -> Optional[str]:
        for spelling in ("dockerfile", "Dockerfile", "DOCKERFILE"):
            if spelling not in os.listdir(project_dir):
                continue

            self.file = spelling

            with open(project_dir / spelling, "r") as file:
                docker_contents = file.readlines()

            return post("parse-container", docker_contents)
        return None


class CIParser:
    def __init__(self) -> None:
        self.ci: List[str] = []

    def parse(self, project_dir: Path) -> str:
        ci_files = {
            "Travis": [".travis.yml", "travis.yml", ".travis.yaml", "travis.yaml"],
            "GitHub Actions": [f".github{os.sep}workflows"],
            "Circle CI": [f".circleci{os.sep}config.yml"],
            "GitLab CI": [".gitlab-ci.yml"],
        }

        for ci_provider, file_names in ci_files.items():
            for file in file_names:
                if (project_dir / file).exists():
                    self.ci.append(ci_provider)
                    break

    def __str__(self) -> str:
        if self.ci:
            if len(self.ci) == 1:
                arch = f"The project uses {self.ci[0]} for CI/CD."
            else:
                arch = (
                    f"The project uses multiple CI/CD providers: {', '.join(self.ci)}"
                )
        else:
            arch = "No CI/CD config files were detected."

        return arch



"""
Detect important environmental packages/libraries.
"""
import os
import re
from pathlib import Path
from typing import List

from jaepeto.utils import post


def read_packages(project_dir: Path) -> List[str]:
    packages = []

    if "requirements.txt" in os.listdir(project_dir):
        with open(project_dir / "requirements.txt", "r") as f:
            contents = f.readlines()

        if new_packages := post("parse-reqs/requirements", contents):
            packages.extend(new_packages)

    for _file in ("environment.yml", "environment.yaml"):
        if _file in os.listdir(project_dir):
            with open(project_dir / _file, "r") as f:
                contents = f.readlines()

            if new_packages := post("parse-reqs/environment", contents):
                packages.extend(new_packages)

    return packages


class PyVersionDetector:
    def __init__(self):
        self.PY37 = 37
        self.PY38 = 38
        self.PY39 = 39
        self.PY310 = 310
        self._version = None
        self.possible_versions = (self.PY38, self.PY39, self.PY310)

        # 3.8
        self._walrus = []
        self._walrus_pattern = re.compile(r"if\s\(?\w*\s?:=\s?.+\)?")

        # 3.9
        self._dict_union = []
        self._dict_union_pattern = re.compile(
            rf"^\s*(?:\w+(?:\s?:\s?\w+(?:\[\w+:\w+\])?)?\s?=\s?)?[\w_]+\s?\|\s?[\w_]+"
        )

        # 3.10
        self._pipe_optional = []
        self._pipe_pattern = re.compile(
            rf":\s?\w+(?:\[\w+:?\w+\])?\s?\|\s?\w+(?:\[\w+:?\w+\])?\s?[,\)=]"
        )

        self._type_hint = []
        self._match_operator = []

    @property
    def version(self) -> str:
        _ver = str(self._version) if self._version else str(self.PY37)
        _ver = _ver[0] + "." + _ver[1:]
        return _ver

    def review(self, filename: str, code_contents: List[str]) -> None:
        __version = self.PY37 if not self._version else self._version

        for considered_version in self.possible_versions:
            if considered_version <= __version:
                continue

            if considered_version == self.PY38:
                self._match_38(filename, code_contents)
            elif considered_version == self.PY39:
                self._match_39(filename, code_contents)
            elif considered_version == self.PY310:
                self._match_310(filename, code_contents)

    def _match_38(self, filename: str, code_contents: List[str]) -> None:
        for _code in code_contents:
            if walrus := self._walrus_pattern.search(_code):
                self._walrus.append(filename)
                self._version = self.PY38

    def _match_39(self, filename: str, code_contents: List[str]) -> None:
        for _code in code_contents:
            if pipe_dict := self._dict_union_pattern.search(_code):
                self._dict_union.append(filename)
                self._version = self.PY39

    def _match_310(self, filename: str, code_contents: List[str]) -> None:
        for _code in code_contents:
            if pipe := self._pipe_pattern.search(_code):
                self._pipe_optional.append(filename)
                self._version = self.PY310

    def __str__(self):
        if not self._version:
            return """
                    No exact compatable Python version has been detected.
                    Versions below 3.8 are compatable.
                    """

        return f"""
                The codebase is compatible with Python {self.version} and below.
                """



from pathlib import Path

from jaepeto.utils import post


def describe_file_architecture(file_path: Path) -> str:
    with open(file_path, "r") as f:
        full_contents = f.read()

    for arch_notifier in ("requests.", "boto3.resource", "MongoClient"):
        if arch_notifier in full_contents:
            with open(file_path, "r") as f:
                contents = f.readlines()

            contents = [c.strip().strip("\n") for c in contents]
            return post("arch", contents)



from pathlib import Path
from typing import Optional

import requests

from jaepeto.utils import post
from jaepeto.utils.parse_python import get_specific_python_function


def summarise_function(function_name: str, file: Path) -> Optional[str]:
    if function_str := get_specific_python_function(function_name, file):
        return post("summarise", function_str.strip("\n"))



"""
Making requests to server
"""
import hashlib
import json
import os
from typing import List, Union

import requests

SERVER_DOMAIN = "https://x0jfh97rfl.execute-api.eu-west-2.amazonaws.com/Prod/"


def post(endpoint: str, payload: Union[str, List[str]]) -> Union[None, str, List[str]]:
    hashed_project = hashlib.sha224(
        os.environ["JAEPETO_PROJECT_NAME"].encode("utf-8")
    ).hexdigest()

    _run_locally = (
        "JAEPETO_RUN_LOCALLY" in os.environ
        and os.environ["JAEPETO_RUN_LOCALLY"] == "True"
    )
    _server_domain = SERVER_DOMAIN if not _run_locally else "http://127.0.0.1:3000/"

    try:
        response = requests.post(
            _server_domain + endpoint + f"?project={hashed_project}",
            data=json.dumps(payload),
            headers={
                "x-api-key": os.environ["JAEPETO_API_KEY"],
                "content-type": "application/json",
            },
        )
    except requests.exceptions.ConnectionError:
        # Happens if running locally but server hasn't set up (or API server has crashed!)
        return None

    if response.status_code == 200:
        message = response.json()["message"]
        if isinstance(message, list):
            return message

        try:
            message = json.loads(message)
        except Exception:
            pass
        return message
    else:
        return None



import ast
from pathlib import Path
from typing import List, Optional, Tuple


def find_func_lines(node: ast.FunctionDef):
    return node.lineno - 1, find_max_func_line(node)


def find_max_func_line(node: ast.FunctionDef):
    max_line = node.lineno

    if hasattr(node, "body"):
        for child in node.body:
            child_max = find_max_func_line(child)
            max_line = max(max_line, child_max)

    if isinstance(node, ast.If):
        for child in node.orelse:
            child_max = find_max_func_line(child)
            max_line = max(max_line, child_max)

    return max_line


def get_specific_python_function(function_name: str, file, max_length: int = 100):
    with open(file, "r") as source:
        tree = ast.parse(source.read())

    for node in tree.body:
        if not isinstance(node, ast.FunctionDef):
            continue

        if node.name == function_name:
            func_start, func_end = find_func_lines(node)
            py_string = _read_python_func(file, func_start, func_end, max_length)

            try:
                ast.parse(py_string)
            except Exception:
                return _read_python_func(file, func_start, func_end + 1, max_length)
            else:
                return py_string


def get_func_lines(file):
    with open(file, "r") as source:
        tree = ast.parse(source.read())

    for node in tree.body:
        if isinstance(node, ast.FunctionDef):
            yield find_func_lines(node)


def get_funcs_as_string(
    file, min_length: int = 6, max_length: int = 100, include_docstring: bool = False
):
    for (func_start, func_end) in get_func_lines(file):
        if func_end - func_start >= min_length:
            func_string = _read_python_func(file, func_start, func_end, max_length)

            if not include_docstring:
                func_string = _remove_docstring(func_string)

            yield func_string


def _remove_docstring(func_string):
    """
    Remove docstring from a function call.

    Notes
    -----
    * Naively searches for triple quotes
    """
    split_func = func_string.split('"""')

    # TODO: Will incorrectly parse code if triple quotes present in code body
    if len(split_func) >= 3:
        split_func[0] = (
            split_func[0].rstrip() + "\n"
        )  # rstrip also removes \n. TODO: use raw strings to avoid this
        split_func[2] = split_func[2].lstrip("\n")

        del split_func[1]

    return "".join(split_func)


def _read_python_func(
    file: Path, min_line: int, max_line: int, max_line_difference: int
) -> str:
    """
    Read a subset of lines in a Python function.

    Parameters
    ----------
    min_line : int
        The starting line number (inclusive) which to read
    max_line : int
        The end line number (inclusive) which to read
    file : pathlib.Path
        The file to read
    max_line_different : int
        Maximum amount of lines to read. Overrides `max_line`.

    Returns
    -------
    str
        Subset of lines from the given file
    """
    with open(file, "r") as source:
        code = source.readlines()

    max_line = min(max_line, min_line + max_line_difference)

    return "".join(code[min_line : max_line + 1])


def identify_main_functions(file: Path):
    with open(file, "r") as f:
        tree = ast.parse(f.read())

    called_funcs: List[str] = []

    for _node in tree.body:
        if not isinstance(_node, ast.If):
            continue

        _node_test = _node.test

        if not isinstance(_node_test, ast.Compare):
            continue

        test_left = _node_test.left

        if not isinstance(test_left, ast.Name):
            # Might be a subscripted comparator, for example
            # Whatever it is, it's certainly not __name__
            continue

        if (
            test_left.id == "__name__"
            and isinstance(_node_test.ops[0], ast.Eq)
            and _node_test.comparators[0].value == "__main__"
        ):
            for sub_node in _node.body:
                if isinstance(sub_node, ast.Expr) or isinstance(sub_node, ast.Assign):
                    sub_node_value = sub_node.value

                    if isinstance(sub_node_value, ast.Call):
                        called_funcs.extend(parse_ast_call(sub_node_value))

    if called_funcs:
        return called_funcs
    else:
        return None


def parse_ast_call(node: ast.Call) -> List[str]:
    called_funcs: List[str] = []

    if _func_call := _parse_ast_func(node.func):
        called_funcs.append(_func_call)

    for _func_arg in node.args:
        if isinstance(_func_arg, ast.Call):
            called_funcs.extend(parse_ast_call(_func_arg))

    for _func_kwarg in node.keywords:
        if isinstance(_func_kwarg.value, ast.Call):
            called_funcs.extend(parse_ast_call(_func_kwarg.value))

    return called_funcs


def _parse_ast_func(node) -> Optional[str]:
    if isinstance(node, ast.Attribute):
        return node.attr
    elif isinstance(node, ast.Name):
        return node.id



"""
Detect important functions to the codebase.
"""
import ast
import os
import re
import subprocess
from itertools import groupby
from operator import itemgetter
from pathlib import Path
from typing import List, Optional, Tuple

from jaepeto.utils.parse_python import (
    get_specific_python_function,
    identify_main_functions,
)

KNOWN_STD_LIB = [
    "__future__",
    "abc",
    "anydbm",
    "argparse",
    "array",
    "asynchat",
    "asyncore",
    "atexit",
    "base64",
    "BaseHTTPServer",
    "bisect",
    "bz2",
    "calendar",
    "cgitb",
    "cmd",
    "codecs",
    "collections",
    "commands",
    "compileall",
    "ConfigParser",
    "contextlib",
    "Cookie",
    "copy",
    "cPickle",
    "cProfile",
    "cStringIO",
    "csv",
    "datetime",
    "dbhash",
    "dbm",
    "decimal",
    "difflib",
    "dircache",
    "dis",
    "doctest",
    "dumbdbm",
    "EasyDialogs",
    "errno",
    "exceptions",
    "filecmp",
    "fileinput",
    "fnmatch",
    "fractions",
    "functools",
    "gc",
    "gdbm",
    "getopt",
    "getpass",
    "gettext",
    "glob",
    "grp",
    "gzip",
    "hashlib",
    "heapq",
    "hmac",
    "imaplib",
    "imp",
    "inspect",
    "itertools",
    "json",
    "linecache",
    "locale",
    "logging",
    "mailbox",
    "math",
    "mhlib",
    "mmap",
    "multiprocessing",
    "operator",
    "optparse",
    "os",
    "pathlib",
    "pdb",
    "pickle",
    "pipes",
    "pkgutil",
    "platform",
    "plistlib",
    "pprint",
    "profile",
    "pstats",
    "pwd",
    "pyclbr",
    "pydoc",
    "Queue",
    "random",
    "re",
    "readline",
    "resource",
    "rlcompleter",
    "robotparser",
    "sched",
    "select",
    "shelve",
    "shlex",
    "shutil",
    "signal",
    "SimpleXMLRPCServer",
    "site",
    "sitecustomize",
    "smtpd",
    "smtplib",
    "socket",
    "SocketServer",
    "sqlite3",
    "string",
    "StringIO",
    "struct",
    "subprocess",
    "sys",
    "sysconfig",
    "tabnanny",
    "tarfile",
    "tempfile",
    "textwrap",
    "threading",
    "time",
    "timeit",
    "trace",
    "traceback",
    "unittest",
    "urllib",
    "urllib2",
    "urlparse",
    "usercustomize",
    "uuid",
    "warnings",
    "weakref",
    "webbrowser",
    "whichdb",
    "xml",
    "xmlrpclib",
    "zipfile",
    "zipimport",
    "zlib",
    "builtins",
    "__builtin__",
]


def identify_module_imports(
    module: str, file_contents: List[str]
) -> List[Tuple[str, str]]:
    imports = []

    for line in file_contents:
        if up_import_match := re.match(r"from \.\.(\w.*) import", line):
            _import_group = up_import_match.group(1)

            if resolved_import := _identify_uplevel_import(
                module, line, _import_group, from_import=True
            ):
                imports.extend(resolved_import)
        elif up_import_match := re.match(r"import \.\.(\w.*)", line):
            _import_group = up_import_match.group(1)

            if resolved_import := _identify_uplevel_import(
                module, line, _import_group, from_import=False
            ):
                imports.extend(resolved_import)
        elif resolved_import := _identify_module_import(module, line):
            imports.extend(resolved_import)

    return imports


def _identify_uplevel_import(
    module: str, line: str, import_group: str, from_import: bool
) -> List[str]:
    if "." in module:
        _new_module = ".".join(module.split(".")[:-1])
    else:
        _new_module = ""

    if from_import:
        _new_line = re.sub(
            rf"from \.\.{import_group} import", rf"from .{import_group} import", line
        )
    else:
        _new_line = re.sub(
            rf"import \.\.{import_group}", rf"import .{import_group}", line
        )

    return _identify_module_import(_new_module, _new_line)


def _identify_module_import(module: str, line: str) -> List[str]:
    imports = []

    if from_match := re.match(
        rf"from ({module}(\..*)?|\..*|[a-zA-Z_]*) import (.*)", line.strip()
    ):
        submodule = from_match.group(1)

        if submodule.startswith("."):
            submodule = module + submodule
        elif module == "" or not submodule.startswith(module):
            if (
                submodule
                in subprocess.check_output(
                    ["pip", "list"], stderr=subprocess.STDOUT
                ).decode("utf-8")
                or submodule in KNOWN_STD_LIB
            ):
                return imports
            else:
                submodule = module + "." + submodule

        for module_import in from_match.group(3).split(","):
            as_split_import = module_import.split("as ")

            import_path = f"{submodule}.{as_split_import[0].strip()}"

            if len(as_split_import) == 1:
                imports.append((import_path, as_split_import[0].strip()))
            elif len(as_split_import) == 2:
                import_representation = as_split_import[1].strip()
                imports.append((import_path, import_representation))
            else:
                raise ImportError(f"Unrecognised import type {module_import}")
    elif import_match := re.match(rf"import ((?:{module}|\.).*)", line.strip()):
        import_path = import_match.group(1).strip()

        if import_path.startswith("."):
            import_path = module + import_path

        imports.append((import_path, import_path))

    return imports


def identify_module_calls(
    imports: List[Tuple[str, str]], file_contents: List[str]
) -> List[str]:
    """
    Identify uses of a module within a codebase.

    Looks for function calls of given imports for a given file.

    Parameters
    ----------
    imports : list of (str, str) tuples
        Module imports in the file. May be imported as a module, or a function using the `from` syntax.
        First element of each tuple is the import. Second element is the imported object as would appear in the file.
    file_contents : list of str
        List of lines in a given file. Lines to read for function calls.

    Returns
    -------
    list of str
        <module>.<function> for each function called in the given file

    Examples
    --------
    >>> imports = [("module.sub", "module.sub")]
    >>> contents = ["result = module.sub.func(1, 2, 3)"]
    >>> identify_module_uses(imports, contents)
        ["module.sub.func"]

    >>> imports = [("module.sub.func", "func_alias")]
    >>> contents = ["func_alias()"]
    >>> identify_module_uses(imports, contents)
        ["module.sub.func"]
    """
    called_imports = []

    for line in file_contents:
        if not imports:
            return []

        for idx, (module_path, module_import) in enumerate(imports):
            if object_call_match := re.search(
                rf"(^|[^a-zA-Z_\.]){module_import}(\..*)?\(.*\)", line
            ):
                if submodule_group := object_call_match.group(2):
                    called_imports.append(module_path + submodule_group)
                else:
                    called_imports.append(module_path)

    return called_imports


def resolve_import_paths(
    module: str, file_path: Path, root: Optional[Path] = None
) -> List[str]:
    """
    Resolve object imports to absolute paths.

    Parameters
    ----------
    module : str
        The name of the module for imports
    file_path : pathlib.Path
        Relative path to a file to read the imports of
    root : pathlib.Path, optional
        Root path from which to form absolute path. If none, use `file_path.parents[1]`.
        Should be one step above `module`.

    Returns
    -------
    list of str
        Absolute imports for functions imported in `file_path`

    Raises
    ------
    RuntimeError
        If a detected imported module does not have an __init__ file and file does not exist
    """
    with open(file_path, "r") as f:
        file_contents = f.readlines()

    absolute_imports: List[str] = []
    imports = identify_module_imports(module, file_contents)

    root = root if root else file_path.parents[1]

    for _import_statement, _ in imports:
        _imported_path = Path(os.sep.join(_import_statement.split(".")[:-1]))
        _imported_path_absolute = root / _imported_path

        if _imported_path_absolute.with_suffix(".py").exists():
            # Import points to file. Assume imported object lies within the file
            absolute_imports.append(_import_statement)
        elif _imported_path_absolute.is_dir():
            _imported_path_init = _imported_path_absolute / "__init__.py"

            if not _imported_path_init.exists():
                raise RuntimeError(
                    f"Trying to resolve {_import_statement}; {_imported_path_init} does not exist"
                )

            if _imported_path_init == file_path:
                # Avoid recursion error from self-referential imports
                continue

            resolved_imports = resolve_import_paths(
                ".".join(_import_statement.split(".")[:-1]),
                _imported_path_init,
                root=root,
            )

            for _resolved_import in resolved_imports:
                # TODO: warn if no import resolution is made
                if _resolved_import.split(".")[-1] == _import_statement.split(".")[-1]:
                    # Assume module does not have multiple objects with same name
                    absolute_imports.append(_resolved_import)
                    break

    return absolute_imports


def infer_important_functions(module: str, root: Path) -> List[str]:
    module_root = root / module

    if (main_file := module_root / "__main__.py").exists():
        # TODO detect called functions in main
        paths = resolve_import_paths(module, main_file)
        if all_imports := detect_all_imports(main_file):
            paths = [p for p in paths if p.split(".")[-1] in all_imports]
        return paths

    elif (init_file := module_root / "__init__.py").exists():
        paths = resolve_import_paths(module, init_file)
        if all_imports := detect_all_imports(init_file):
            paths = [p for p in paths if p.split(".")[-1] in all_imports]
        return paths

    else:
        return []


def detect_all_imports(file: Path):
    with open(file, "r") as f:
        tree = ast.parse(f.read())

    for elem in tree.body:
        if not isinstance(elem, ast.Assign):
            continue

        if elem.targets[0].id != "__all__":
            continue

        return [all_import.value for all_import in elem.value.elts]

    return None


def group_imports_at_level(imports: List[str], level: int):
    if level < 1:
        raise ValueError(f"Level must be greater than 0, was {level}")

    short_imports = [
        ".".join(split_import[:level])
        for split_import in [_import.split(".") for _import in imports]
        if len(split_import) >= level
    ]
    short_imports.sort()

    grouped_imports = []

    for _import, grouper in groupby(short_imports):
        counts = 0
        for _ in grouper:
            counts += 1
        grouped_imports.append((_import, counts))

    grouped_imports.sort(key=itemgetter(1), reverse=True)
    return grouped_imports


def extract_main_calls(py_file: Path, module: str, root: Optional[Path] = None):
    main_calls = identify_main_functions(py_file)

    if not main_calls:
        return None

    funcs = []
    func_names = []
    for _func_name in main_calls:
        if func_string := get_specific_python_function(_func_name, py_file):
            funcs.append(func_string)
            func_names.append(module + "." + py_file.stem + "." + _func_name)

    for _func_import in resolve_import_paths(py_file.parents[0].stem, py_file):
        _func_name = _func_import.split(".")[-1]

        if _func_name in main_calls and _func_name not in func_names:
            _func_path = os.sep.join(_func_import.split(".")[:-1]) + ".py"
            if root:
                _func_path = root / _func_path
            if os.path.exists(_func_path):
                if _func_string := get_specific_python_function(_func_name, _func_path):
                    funcs.append(_func_string)
                    func_names.append(_func_import)

    if funcs:
        return [(_name, _func) for _name, _func in zip(func_names, funcs)]
    else:
        return None


def get_script_entrypoints(module_root: Path):
    root = module_root.parent

    func_entrypoints = {}

    for py_file in module_root.glob("*.py"):
        # Going deeper gets too complicated at the moment.
        _file_path = f"{module_root.stem}.{py_file.stem}"

        if _main_calls := extract_main_calls(py_file, module_root.stem, root):
            _main_funcs = [f[0] for f in _main_calls]
            func_entrypoints[_file_path] = _main_funcs

    return func_entrypoints



"""
Parse user config
"""
import os
from configparser import ConfigParser, NoOptionError
from pathlib import Path
from typing import Union


def _clean_config_dir(value: str) -> str:
    """
    Remove errant / and . from directory values in config.

    Leading `./` and trailing `/`
    are to be removed as the code does not expect paths
    in this format and consequently cannot resolve the paths.

    Parameters
    ----------
    value : str
        The unclean directory

    Returns
    -------
    str
        A clean directory
    """
    value = value.lstrip("./")
    value = value.rstrip("/")
    return value


def parse_config(config_file: Union[Path, str]):
    config = ConfigParser()
    config.read(config_file)

    project_name = config.get("project", "name")
    project_name = project_name if project_name else "Project"

    project_info = {
        "name": project_name,
        "description": config.get("project", "description"),
        "src": _clean_config_dir(config.get("project", "srcdir")),
        "doc_dir": _clean_config_dir(config.get("docs", "docdir")),
    }

    try:
        local = config.getboolean("project", "local")
    except NoOptionError:
        local = False

    os.environ["JAEPETO_API_KEY"] = config.get("project", "api_key")
    os.environ["JAEPETO_PROJECT_NAME"] = project_name
    os.environ["JAEPETO_RUN_LOCALLY"] = str(local)

    language = config.get("project", "language")

    if language.lower().strip().rstrip("3") != "python":
        raise RuntimeError(
            f"""Jaepeto currently only works
        on Python (3), not {language}"""
        )

    # Documentation
    project_info["docs"] = {"technical": True, "technical_name": "technical_doc"}

    # Issues
    if config.has_section("issues"):
        issue_service = config.get("issues", "integration")
        if issue_service and issue_service.lower() != "github":
            project_info["issues"] = {}
        else:
            try:
                username = config.get("issues", "username")
                repo = config.get("issues", "repo")
            except:
                project_info["issues"] = {}
            else:
                project_info["issues"] = {"username": username, "repo": repo}
    else:
        project_info["issues"] = {}

    return project_info


def create_config():
    """
    Create a minimal config file.

    Must be called when project root is current working directory.
    Some information must be filled out by a user, such as `api_key`.
    `srcdir` is assumed to be `src` or just home directory.

    Notes
    -----
    * Writes config file to CWD / ".jaepeto.ini"
    """
    project = Path.cwd()
    config_path = project / ".jaepeto.ini"

    if config_path.exists():
        print(".jaepeto.ini already exists!")
        return

    config = ConfigParser()

    project_name = project.stem

    if (project / "src").is_dir():
        srcdir = "src"
    else:
        srcdir = "."

    config.add_section("project")
    config.set("project", "name", project_name)
    config.set("project", "description", "<ENTER SHORT PROJECT DESCRIPTION>")
    config.set("project", "language", "python")
    config.set("project", "api_key", "<ADD API KEY HERE>")
    config.set("project", "srcdir", srcdir)

    config.add_section("docs")
    config.set("docs", "docdir", "./")

    print(f"Creating config file at {config_path}")
    with open(config_path, "w") as f:
        config.write(f)



import re
import subprocess
from collections import defaultdict
from datetime import datetime

leading_4_spaces = re.compile("^    ")


def get_commits():
    lines = (
        subprocess.check_output(["git", "log", "--stat"], stderr=subprocess.STDOUT)
        .decode("utf-8")
        .split("\n")
    )

    commits = []
    current_commit = {}

    def save_current_commit():
        title = current_commit["message"][0]
        message = current_commit["message"][1:]
        changed_files = []

        if message and message[0] == "":
            del message[0]

        current_commit["title"] = title
        current_commit["message"] = "\n".join(message)
        current_commit["files"] = [
            _m.split("|")[0].strip() for _m in message if "|" in _m
        ]
        commits.append(current_commit)

    for line in lines:
        if not line.startswith(" "):
            if line.startswith("commit "):
                if current_commit:
                    save_current_commit()
                    current_commit = {}
                current_commit["hash"] = line.split("commit ")[1]
            else:
                try:
                    key, value = line.split(":", 1)
                    current_commit[key.lower()] = value.strip()
                except ValueError:
                    pass
        else:
            current_commit.setdefault("message", []).append(
                leading_4_spaces.sub("", line)
            )

    if current_commit:
        save_current_commit()

    commits.reverse()  # earliest commit first
    return commits


def find_earliest_commit_with_file(file: str, commits):
    # Assume commits are ordered earliest to latest
    for commit in commits:
        for commit_file in commit["files"]:
            if file in commit_file.lower():
                # Accepts partial match e.g. .travis -> .travis.yml
                return convert_datetime_to_date_string(commit["date"])

    raise RuntimeError(f"No commits involve file {file}")


def get_diff_for_file(file: str, commits):
    changes = []

    for i in range(1, len(commits)):
        main_commit = commits[i]

        if file not in main_commit["files"]:
            continue

        previous_commit_hash = commits[i - 1]["hash"]
        main_commit_hash = main_commit["hash"]

        lines = (
            subprocess.check_output(
                ["git", "diff", previous_commit_hash, main_commit_hash, file],
                stderr=subprocess.STDOUT,
            )
            .decode("utf-8")
            .split("\n")
        )

        lines = [l for l in lines if re.match("[+-][a-zA-Z]", l)]

        formatted_date = convert_datetime_to_date_string(main_commit["date"])
        changes.append((main_commit_hash, formatted_date, lines))

    return changes


def convert_datetime_to_date_string(datetime_string):
    return datetime.strptime(datetime_string, "%a %b %d %H:%M:%S %Y %z").strftime(
        "%Y-%m-%d"
    )


def get_line_changes_for_file(file: str):
    lines = (
        subprocess.check_output(
            ["git", "log", "-p", '--since="4 weeks ago"', "--", file],
            stderr=subprocess.STDOUT,
        )
        .decode("utf-8")
        .split("\n")
    )

    commits = []
    current_commit = {}

    def _save_current_commit():
        title = current_commit["message"][0]
        message = current_commit["message"][1:]
        changed_files = []

        if message and message[0] == "":
            del message[0]

        current_commit["title"] = title
        commit_message = "\n".join(message)

        if "additions" in current_commit:
            lines_added = len(current_commit["additions"])
        else:
            lines_added = 0

        if "deletions" in current_commit:
            lines_removed = len(current_commit["deletions"])
        else:
            lines_removed = 0

        commits.append((lines_added, lines_removed, current_commit["author"]))

    for line in lines:
        if not line.startswith(" "):
            if line.startswith("commit "):
                if current_commit:
                    _save_current_commit()
                    current_commit = {}
                current_commit["hash"] = line.split("commit ")[1]
            else:
                if line == "+" or re.match(r"\+[^\+]", line):
                    current_commit.setdefault("additions", []).append(line)
                elif line == "-" or re.match(r"-[^-]", line):
                    current_commit.setdefault("deletions", []).append(line)
                else:
                    try:
                        key, value = line.split(":", 1)
                        current_commit[key.lower()] = value.strip()
                    except ValueError:
                        pass
        else:
            current_commit.setdefault("message", []).append(
                leading_4_spaces.sub("", line)
            )

    if current_commit:
        _save_current_commit()

    authors = {}
    for _added, _removed, _author in commits:
        authors.setdefault(_author, 0)
        authors[_author] += _added + _removed

    if authors:
        common_author = max(authors, key=authors.get)
    else:
        common_author = None

    return sum([c[0] for c in commits]), sum([c[1] for c in commits]), common_author



import re

from setuptools import find_packages, setup


def get_version():
    result = re.search(
        r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]', open("jaepeto/__init__.py").read()
    )
    return result.group(1)


setup(
    version=get_version(),
)



"""
Write technical docs
"""
import os
import re
from datetime import datetime
from pathlib import Path
from textwrap import dedent
from typing import Optional

from jaepeto.patterns import (
    CIParser,
    ContainerParser,
    PyVersionDetector,
    describe_file_architecture,
    get_script_entrypoints,
    group_imports_at_level,
    infer_important_functions,
    read_packages,
)
from jaepeto.summaries import summarise_function
from jaepeto.utils import get_line_changes_for_file, parse_config, post


def create_technical_doc(
    source_module_name: str,
    source_code_dir: Path,
    technical_doc_file: Path,
    repo_info: str = None,
    project_title: str = "Project",
    project_summary: Optional[str] = None,
):
    write_document_intro(technical_doc_file, project_title, project_summary)
    create_architecture_section(source_code_dir, technical_doc_file)
    create_src_overview(
        source_module_name, source_code_dir / source_module_name, technical_doc_file
    )
    create_module_overview(source_code_dir, technical_doc_file)


def write_document_intro(technical_doc_file: Path, project_title: str, project_summary):
    with open(technical_doc_file, "w") as file:
        file.write(f"# {(project_title + ' Technical Documentation').strip()}")
        file.write(f"\n\n**Last updated:** {datetime.now().strftime('%Y-%m-%d')}\\")
        file.write("\n_Document generation aided by **Jaepeto**_")

        if project_summary:
            file.write(f"\n\n{project_summary}")

        file.write("\n\n* [Introduction](#introduction)")
        file.write("\n* [Code Overview](#code-overview)")

        file.write("\n\n## Introduction")
        file.write(
            f"""\n\nThis is a technical document detailing
        at a high-level
        what {project_title} does, how it operates,
        and how it is built.
        It is meant for consumption internally
        by technical developers
        and managers."""
        )

        file.write("\n\n### Scope")
        file.write(
            """\n\nThis document provides high-level summaries
        of the project codebase."""
        )

        file.write("\n\n### Context")
        file.write(
            """\n\nThe outline of this document was generated
        by **Jaepeto**.
        The documentation should be reviewed
        and detail provided manually
        before the documentation is released."""
        )


def create_architecture_section(source_code_dir: Path, technical_doc_file: Path):
    with open(technical_doc_file, "r") as f:
        contents = f.readlines()

    contents.append("\n\n## Project Architecture\n\n")

    if packages := read_packages(source_code_dir):
        contents.append("The project has the following dependencies:\n")
        for _package in packages:
            contents.append(f"\n* {_package}")
        contents.append("\n\n")

    container_parser = ContainerParser()
    if container_arch := container_parser.parse(source_code_dir):
        contents.append(container_arch + "\n\n")

    ci_parser = CIParser()
    ci_parser.parse(source_code_dir)
    if ci_arch := str(ci_parser):
        contents.append(ci_arch + "\n\n")

    with open(technical_doc_file, "w") as file:
        file.writelines(contents)


def create_src_overview(
    module_name: str, source_code_dir: Path, technical_doc_file: Path
):
    with open(technical_doc_file, "r") as file:
        contents = file.readlines()

    # Assume contents have intro
    contents.append("\n\n## Code Overview\n\n")

    # Detect Python version
    version_detector = PyVersionDetector()

    for pyfile in source_code_dir.glob("**/*.py"):
        with open(pyfile, "r") as f:
            _pyfile_contents = f.readlines()

        version_detector.review(pyfile, _pyfile_contents)
    del _pyfile_contents
    contents.append(str(version_detector) + "\n\n")

    # Important / __main__/__init__ functions
    entrypoints = infer_important_functions(module_name, source_code_dir.parent)
    num_entrypoints = len(entrypoints)

    if 0 < num_entrypoints <= 5:
        contents.append(
            f"There are {num_entrypoints} source code objects in top-level `__main__`/`__init__`:\n\n"
        )
        for _entrypoint in entrypoints:
            entrypoint_func = _entrypoint.split(".")[-1]
            entrypoint_file = (os.sep).join(_entrypoint.split(".")[:-1]) + ".py"
            contents.append(f"### `{entrypoint_func}` from `{entrypoint_file}`\n")

            _entrypoint_doc = summarise_function(entrypoint_func, entrypoint_file)
            if _entrypoint_doc:
                contents.append(f" \n```\n{_entrypoint_doc}\n```\n")
            else:
                contents.append(
                    "\nFunction cannot be summarised: No documentation is present for this code.\n"
                )

        contents.append(
            "\nThese entrypoints are broken down into the following modules:\n\n"
        )
    elif num_entrypoints == 0:
        contents.append(
            f"There are 0 source code entrypoints in top-level `__main__`/`__init__` files.\n"
        )
    else:
        contents.append(
            f"""There are {num_entrypoints} source code objects
                        in top-level `__main__`/`__init__`.
                        They are broken down into the following modules:\n\n"""
        )

    entrypoint_groups = group_imports_at_level(entrypoints, level=2)
    for _entrypoint, _count in entrypoint_groups:
        contents.append(f"* `{_entrypoint}` has {_count} entrypoints\n")

    contents.append("\n")
    # Scripts with __name__ == __main__ and function calls
    script_entrypoints = get_script_entrypoints(source_code_dir)
    for _file, _entrypoints in script_entrypoints.items():
        if not _entrypoints:
            contents.append(f"`{_file}`` has a `__main__` entrypoint.\n")
        else:
            contents.append(f"`{_file}` has a `__main__` entrypoint, which calls:\n\n")
            for _entrypoint in _entrypoints:
                contents.append(f"* `{_entrypoint}`\n")

    with open(technical_doc_file, "w") as file:
        file.writelines(contents)


def create_module_overview(source_code_dir: Path, technical_doc_file: Path):
    """
    Create an overview of Python file structure in the project.
    """
    doc_contents, _ = _create_module_overview(
        source_code_dir, [f"\n### **{source_code_dir.stem}/**"], 0
    )

    if len(doc_contents) > 1:
        # if == 1, then only header was present but no python files. Therefore exclude.
        with open(technical_doc_file, "a") as file:
            file.writelines(doc_contents)


def _create_module_overview(module_path: Path, doc_contents, n_calls: int):
    """
    Create an overview of Python file structure in the project.

    Include directories and subdirectories if they contain at least 1 python file.
    For each file, describe the rate of the change of the file,
    an inferred owner, and consequences of the file such as API or database calls.

    Parameters
    ----------
    module_path : pathlib.Path
        The path to the current directory being explored
    doc_contents : list of str
        List containing contents above in the directory structure
        to be added to documentation
    n_calls : int
        DEPRECATED. Number of calls to API in this module overview section

    Returns
    -------
    list or str
        The contents to add to the documentation
    int
        DEPRECATED. The total number of API calls in this section
    """
    hash_level = min(6, len(doc_contents[-1].replace("\n", "").split()[0]) + 1)

    for subdir in module_path.glob("*"):
        if subdir.stem == "__pycache__":
            continue

        if subdir.is_dir():
            new_doc_contents, n_calls = _create_module_overview(
                subdir, [f"\n\n{'#'*hash_level} {subdir.stem}/"], n_calls
            )
            if len(new_doc_contents) > 1:
                doc_contents.extend(new_doc_contents)

    for py_file in module_path.glob("*.py"):
        if py_file.stem == "__init__":
            continue

        doc_contents.append(f"\n\n{'#'*hash_level} {py_file.stem}.py")

        lines_added, lines_removed, most_common_author = get_line_changes_for_file(
            str(py_file)
        )
        doc_contents.append(
            dedent(
                f"""
            \n\nFile has {lines_added} lines added and {lines_removed} lines removed
            in the past 4 weeks."""
            )
        )
        if most_common_author:
            doc_contents[-1] += f" {most_common_author} is the inferred code owner."

        doc_contents.append("\n")

        if file_arch := describe_file_architecture(py_file):
            doc_contents.append(file_arch)

    return doc_contents, n_calls


def generate_documentation() -> None:
    """
    Generate and save technical documentation.

    This is a CLI entrypoint to the codebase.
    The current working directory must be the root of the project
    for which documentation is to be generated.
    The root project must contain a `.jaepeto.ini` file.
    """
    config_path = Path.cwd() / ".jaepeto.ini"

    if not config_path.exists():
        print(f"Error running jaepeto: No config file found at {config_path}")
    else:
        config = parse_config(config_path)

        tech_doc_path = (
            config_path.parent / config["doc_dir"] / config["docs"]["technical_name"]
        ).with_suffix(".md")

        post("notify", None)

        create_technical_doc(
            config["src"],
            config_path.parent,
            tech_doc_path,
            project_title=config["name"],
            project_summary=config["description"],
        )
        print(f"Successfully generated technical documentation at {tech_doc_path}")
